/*
 * inth.c  -dl
 * Main file for inth (IP NAT to Host) utility.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>

#include "packet_headers.h"
#include "network.h"
#include "utility.h"
#include "inth.h"

void inth_help()
{
	char help[] =
	"IP NAT to Host (inth) Utility - Help Text.\n"
	"\n"
	"Usage: ./inth -i [iface] -d [destination] -p [ports]\n"
	"\n"
	"Options:\n"
	"  -i [interface] Network Interface\n"
	"  -h : Display this help text\n"
	"  -v : Enable verbose output\n"
	"  -c : Surpress Output from closed ports\n"
	"  -t [secs] Time to wait for response\n"
	"  -d [Destination address]\n"
	"  -p [Port numbers] Comma seperated list '21,22,80' or seqential list '80-100'\n"
	//"  -z : Attempt to fingerprint NAT device if NAT is suspected\n"
	//"  -a : Attempt to detect Asymmetric routing\n"
	"\n"
	"Sig format:[Win:ttl:mss:df:wscale:sackok:nop:size:type]\n";
	
	printf(help);
}

/*
 * Parse in sig database, this is the bare minimum for functionality,
 * much work must be done here, returns number of sigs read on success 
 * or -1 on failure.
*/
int read_sigdb(char *file, fingerp_t *sigDB)
{
	int i,x,y,z;
	FILE *rfd;
	char buf[MAXL],*line, tmp[MAXL], *nptr, os[MAXL];
	u_short num;
	
	/*Open file*/
	if(!(rfd = fopen("./disco.fp", "r"))){
		perror("fopen");
		return -1;
	}
	
	/* Scan in database */
	z=0;
	while((line = fgets(buf, MAXL, rfd)) != NULL){
		if(line[0] == '#' || line[0] == '\n')
			continue;
		if((strstr(line, ":A:") == NULL) && (strstr(line, ":R:") == NULL))
			continue;
		if(z >= MAXSIGS){
			printf("# Sig database is too large, please adjust the MAXSIGS variable "
				   "in inth.h.\n# Currently working off %d sigs.\n\n", z);
			return z;
		}
		
		/* Parse manually to avoid sscanf()/strtok() */
		i = 0; x = 0, y = 0;
		while(line[i] != '\n'){
			tmp[x++] = line[i];
			if((line[i] == ':') && 
				(isdigit(line[i-1]) != 0) && y <= 9){
				tmp[x] = '\0';
				num = atoi(tmp);  
				if(y == 0) sigDB[z].win = num;
				else if(y == 1) sigDB[z].ttl = num;
				else if(y == 2) sigDB[z].mss = num;
				else if(y == 3) sigDB[z].df = num;
				else if(y == 4) sigDB[z].wscale = (num == 65535) ? -1 : num;
				else if(y == 5) sigDB[z].sackok = num;
				else if(y == 6) sigDB[z].nop = num;
				else if(y == 7) sigDB[z].psize = (num == 65535) ? -1 : num;
				x=0; y++;
			}
			
			/* Not interested in SYN sigs, only ACK & RST*/
			if((line[i] == 'A' || line[i] == 'R') && 
			   (line[i-1] == ':' && line[i+1] == ':')){
				strncpy(&sigDB[z].ptype, (char*)tmp, 1);
				/* Grab OS string */
				if((nptr = strrchr(line, ':')) != NULL){
					strcpy(os, nptr+1);
					chomp(os);
					strcpy(sigDB[z].os, os);
				}
				
				x=0; y++;
			}
			
			i++;
		}
		/*TODO: Validate sig */

		/* DBG
		printf("%d:%d:%d:%d:%d:%d:%d:%d:%c:%s",sigDB[z].win,sigDB[z].ttl,sigDB[z].mss,
					sigDB[z].df,sigDB[z].wscale,sigDB[z].sackok,sigDB[z].nop, sigDB[z].psize,
					sigDB[z].ptype, sigDB[z].os);
		*/
	
		z++;	
	}

	fclose(rfd);
	return z;
}

/*
 * Filters response TCP packet from the target, using IP addresses and
 * port numbers, returns 0 on success or -1 on failure
*/
int filter_tcp_packet(char *packet, int rsize, ipcfg_t ip_cfg, tcpcfg_t tcp_cfg)
{
	iphdr_t *ip_hdr = xmalloc(sizeof(iphdr_t));
	tcphdr_t *tcp_hdr = xmalloc(sizeof(tcphdr_t));
	
	/* Check packet is large enough to be valid TCP */
	if(rsize < (sizeof(iphdr_t) + sizeof(tcphdr_t)))
		return -1;
	
	/* Check Protocol type and IP addressess */
	ip_hdr = (iphdr_t *)packet;
	tcp_hdr = (tcphdr_t *)(packet + ip_hdr->ip_hl * 4);
	if((ip_hdr->ip_p == IPPROTO_TCP) &&
		(memcmp(&ip_hdr->ip_dst, &ip_cfg.src, IP_ALEN) == 0) &&
		(memcmp(&ip_hdr->ip_src, &ip_cfg.dest, IP_ALEN)) == 0){
		
		/* Check for source and dest ports */
		if((ntohs(tcp_hdr->th_sport) == tcp_cfg.dport) &&
			(ntohs(tcp_hdr->th_dport) == tcp_cfg.sport)){
				return 0;
		}
	} 
	
	/* Packet not ours */
	return -1;
}

/*
 * Extract fingerprint from TCP packet, this is currently rather
 * primative (old p0f sigs), future versions will attempt to cover many 
 * of M Z's new tests and quirks from p0f v2, especially multiple NOPS 
 * and their order in the options, returns fingerp_t
*/
fingerp_t *extract_fprint(char *packet)
{
	int i, optsz,topt;
	char *ptr;
	fingerp_t *fingerp = xmalloc(sizeof(fingerp_t));
	iphdr_t *ip_hdr = xmalloc(sizeof(iphdr_t));
	tcphdr_t *tcp_hdr = xmalloc(sizeof(tcphdr_t));
	
	ip_hdr = (iphdr_t *)packet;
	tcp_hdr = (tcphdr_t *)(packet + ip_hdr->ip_hl * 4);
	
	/* Grab standard parameters */
	fingerp->ttl = ip_hdr->ip_ttl;
	fingerp->df = (ntohs(ip_hdr->ip_off) == IP_DF) ? 1 : 0;
	fingerp->win = ntohs(tcp_hdr->th_win);
	
	/* Grab Packet size */
	fingerp->psize = (ip_hdr->ip_hl * 4) + (tcp_hdr->th_off * 4);
	
	/* Initiaize options fields */
	fingerp->mss = 0;
	fingerp->wscale = 0;
	fingerp->sackok = 0;
	fingerp->nop = 0;
	
	/* Do we have any options? */
	if((tcp_hdr->th_off * 4) > 20){
		/* How many? */
		fingerp->topts =  (fingerp->psize - ((ip_hdr->ip_hl * 4)+20)) / 4;
		optsz = fingerp->topts * 4;
		
		/* find start of options */
		ptr = (char*)packet + (fingerp->psize - optsz);
		
		/* Parse options */
		i = 0;
		while(i < optsz){
			topt = (*(ptr+i));
			
			if(topt == TCPOPT_MAXSEG){
				fingerp->mss = GET16(ptr+i+2);
				i+=4; 
			}
			else if(topt == TCPOPT_WINDOW){
				fingerp->wscale = (*(ptr+i+2));
				i+=3; 
			}
			else if(topt == TCPOPT_SACK_PERMITTED){
				fingerp->sackok = 1;
				i+=2; 
			}
			else if(topt == TCPOPT_NOP){
				fingerp->nop = 1;
				i++;
			}
			/* End of Options */
			else if(topt == TCPOPT_EOL){
				i = optsz;
			}
			else { i++;}
		}
	}
	
	/* Grab packet type */
	if(tcp_hdr->th_flags == TH_SYN+TH_ACK) strncpy(&fingerp->ptype, "A", 1);
	if(tcp_hdr->th_flags == TH_RST ||
		tcp_hdr->th_flags == TH_ACK+TH_RST) strncpy(&fingerp->ptype, "R", 1);

	
	return (fingerp);
}

/*
 * Searches sigDB for possible matches
 *
*/
int search_siglist(fingerp_t *fingerp, fingerp_t *sigDB, int nsigs, 
				   outcntl_t out, ipcfg_t ip_cfg, tcpcfg_t tcp_cfg)
{
	int z, x;
	int sigin[5];
	
	/* DBG
	printf("%d:%d:%d:%d:%d:%d:%d:%d\n",fingerp->win,fingerp->ttl,fingerp->mss,
					fingerp->df,fingerp->wscale,fingerp->sackok,fingerp->nop, fingerp->psize);
	*/
	
	/* Only search on SYN-ACKS */
	x = 0;
	if(fingerp->ptype == 'A'){
		z = 0; 
		while(z < nsigs){
			if(fingerp->win != sigDB[z].win){ 
				z++; 
				continue; 
			}
			if(((sigDB[z].ttl - fingerp->ttl) > 30 ) ||
			   (sigDB[z].ttl < fingerp->ttl)){ 
		  		z++; 
				continue; 
			}
			if(fingerp->mss != sigDB[z].mss){ 
				z++; 
				continue; 
			}
			if(fingerp->df != sigDB[z].df){ 
				z++; 
				continue; 
			}
			if((sigDB[z].wscale != -1) && 
			   (fingerp->wscale != sigDB[z].wscale)){ 
				z++; 
				continue; 
			}
			if(fingerp->sackok != sigDB[z].sackok){ 
				z++; 
				continue; 
			}
			if(fingerp->nop != sigDB[z].nop){
				z++; 
				continue; 
			}
			if((sigDB[z].psize != -1) && 
			   (fingerp->psize != sigDB[z].psize)){ 
				z++; 
				continue; 
			}
			
			
			if(!x) printf("%s:[%d] ", inet_ntoa(ip_cfg.dest), tcp_cfg.dport);
			if(!x) printf("%s",sigDB[z].os);
			if(x) printf("| %s",sigDB[z].os);
			
			sigin[x] = z;
			x++;
			
		z++;
		}
		
		if(!out.verbose && x) printf("\n");
		if(out.verbose && x) printf(" Signature:[%d:%d:%d:%d:%d:%d:%d:%d:%c]\n",
						fingerp->win,fingerp->ttl,fingerp->mss,fingerp->df,
						fingerp->wscale,fingerp->sackok,fingerp->nop, 
						fingerp->psize,fingerp->ptype);
	}
	
	/* If we have no match */
	if(!x){
		if(fingerp->ptype == 'R' && !out.sclosed){
			printf("%s:[%d] ", inet_ntoa(ip_cfg.dest), tcp_cfg.dport);
			if(out.verbose) printf("Signature:[%d:%d:%d:%d:%d:%d:%d:%d:%c]\t",
								fingerp->win,fingerp->ttl,fingerp->mss,fingerp->df,
			   					fingerp->wscale,fingerp->sackok,fingerp->nop, 
			   					fingerp->psize, fingerp->ptype);
		 	printf(" (Port closed)\n");
		}
		if(fingerp->ptype == 'A'){
			printf("%s:[%d] {Unknown}", inet_ntoa(ip_cfg.dest), tcp_cfg.dport);
			if(out.verbose) printf("Signature:[%d:%d:%d:%d:%d:%d:%d:%d:%c]\t",
								fingerp->win,fingerp->ttl,fingerp->mss,fingerp->df,
			   					fingerp->wscale,fingerp->sackok,fingerp->nop, 
			   					fingerp->psize, fingerp->ptype);
			printf("\n");
		}	
	}
	
	
	return 0;
}

int main(int argc, char **argv)
{
	int opts, RDNS = 0, nports, cntl =0, rsize, gotpk;
	outcntl_t out;
	out.verbose = 0, out.sclosed = 0;
	
	/* Network Sockets */
	/* Currently using a socket each for send and receive as
	   PF_PACKET, SOCK_DGRAM seems to not find the local destination
	   MAC. */
	int sendsk, recvsk;
	
	/* Interface config */
	ifcfg_t *if_cfg;
	
	/* Packet Headers */
	iphdr_t *ip_hdr;
	tcphdr_t *tcp_hdr;
	
	/* Packet configs */
	ipcfg_t ip_cfg;
	tcpcfg_t tcp_cfg;
	u_short *ports;
	
	/* Timout */
	u_int32_t start;
	int timeout = 2;
	
	/* Packets */
	char *packet, *rpacket;
	char *payload = NULL;
	int pksize;
	
	/* OS fingerprints */
	fingerp_t *fingerp;
	
	/* SigDB */
	fingerp_t sigDB[MAXSIGS];
	int nsigs;
	
	/* Check for admin rights */
	if(getuid() != 0 || getgid() != 0){
		printf("Root privilages needed to run this application.\n");
		exit (-1);
	}
	
	if_cfg = xmalloc(sizeof(ifcfg_t));
	/* Grab User Options */
	while((opts = getopt(argc, argv, "i:hvVcnt:d:p:")) != -1){
		switch(opts){
			case 'i':
				if_cfg->iface = xmalloc(strlen(optarg)+1);
				strcpy(if_cfg->iface, optarg);
				break;
			case 'h':
				inth_help();
				return 0;
			case 'v':
				out.verbose = 1;	
				break;
			case 'V':
				printf("IP/Nat/Host Fingerprinting Utility [inth v%s]\n", 
					   VERSION);
				return 0;
			case 'c':
				out.sclosed = 1;
				break;
			case 'n':
				RDNS = 1;
				break;
			case 't':
				timeout = atoi(optarg);
				if(timeout < 1 || timeout > 15){
					printf("Invalid timeout, [1-15].\n");
					return -1;
				}
				break;
			case 'd':
				if((resolve_name(optarg, &ip_cfg.dest)) != 0){
					printf("Invalid destination address supplied.\n");
					return -1;
				}
				break;
			case 'p':
				ports = xmalloc(sizeof(u_short));
				if((strrchr(optarg, ',')==NULL) && (strrchr(optarg, '-')==NULL)){
					ports[0] = atoi(optarg);
					nports = 1;
					break;
				}
				if((nports = handle_ports(optarg, ports)) == -1){
					printf("Incorrect ports format supplied\n");
					return -1;
				}
				break;
			default:
				printf("Usage: ./inth -i [iface] -d [destination] -p [ports]\n");
				return -1;
		}
	}
	
	/* Basic Checks */
	if(!if_cfg->iface){
		printf("No interface supplied.\n");
		return -1;
	}
	
	/* Open network sockets */
	sendsk = open_raw_inet(IPPROTO_RAW);
	recvsk = open_raw_pf(htons(ETH_P_IP));
	
	/* Make receive socket non-blocking*/
	int flags = fcntl(recvsk, F_GETFL);
	flags |= O_NONBLOCK;
	fcntl(recvsk, F_SETFL, flags);

	
	/* Grab local Iface info */
	if((get_local_info(sendsk, if_cfg)) != 0){
		printf("Unable to get local network device info\n");
		return -1;
	}
	
	/* Bind receive socket to network interface*/
	if((bind_rsock2dev(if_cfg->iface, recvsk, ETH_P_IP)) != 0){
		printf("Unable to bind to device\n");
		return -1;	
	}
	
	/* Read in OS sig Database */
	if((nsigs = read_sigdb("G", sigDB)) == -1)
		return -1;
	
	/* Verbose Information */
	if(out.verbose)
		printf("Probing Address %s\n\n", inet_ntoa(ip_cfg.dest));
	
	/* Build Initial Protocol Configuration */
	srand(time(NULL));
	ip_cfg.ip_id = 1 + (u_short)(65535.0 * (rand() / (RAND_MAX + 100.0)));
	ip_cfg.frag_off = IP_DF;  //Mimic Linux2.6x, DF SET
	ip_cfg.ttl = 64;
	ip_cfg.ipp = IPPROTO_TCP;
	memcpy(&ip_cfg.src, &if_cfg->ipa, 4);
	
	tcp_cfg.seqn = 1 + (u_int32_t)(65535.0 * (rand() / (RAND_MAX + 1.0)));
	tcp_cfg.ackn = 0;
	tcp_cfg.sport = 1 + (u_short)(65535.0 * (rand() / (RAND_MAX + 1024.0)));
	tcp_cfg.tflags = TH_SYN;
	tcp_cfg.winsz = 1460*4; //Mimic Linux2.6x, MMS * 4
	tcp_cfg.urgp = 0;
	
	/* Build TCP Options */
	tcp_cfg.optsz = 20;
	tcp_cfg.tcpopts = xmalloc(tcp_cfg.optsz);
	memcpy(tcp_cfg.tcpopts, TCPOPTS, 8);
	
	/* Using time for TSVal as system uptime provides a little too much 
	 * information about the sending host for my liking */
	u_int32_t t = htonl(time(NULL));
	memcpy(tcp_cfg.tcpopts+8, &t, 4);
	memcpy(tcp_cfg.tcpopts+8+4, TCPOPT2, 8);
	
	/* Main Control Loop */
	while(cntl < nports){ 
		
		ip_cfg.ip_id -= 2;
		tcp_cfg.seqn +=1;
		tcp_cfg.sport = (tcp_cfg.sport<65470) ? tcp_cfg.sport+ip_cfg.ttl : 2048;
		tcp_cfg.dport = ports[cntl++];
		
		/* Build packet */
		ip_hdr = build_ip_hdr(ip_cfg, sizeof(tcphdr_t), 0);
		tcp_hdr = build_tcp_hdr(tcp_cfg, ip_hdr, payload ,0);
		pksize = sizeof(iphdr_t) + sizeof(tcphdr_t) + tcp_cfg.optsz;
		packet = xmalloc(pksize);
		memcpy(packet, ip_hdr, sizeof(iphdr_t));
		memcpy(packet + sizeof(iphdr_t), tcp_hdr, sizeof(tcphdr_t));
		memcpy(packet + sizeof(iphdr_t) + sizeof(tcphdr_t), tcp_cfg.tcpopts, tcp_cfg.optsz);
	
		/* Fire Packet on socket */
		if((send_raw_inet(sendsk, packet, pksize, ip_cfg.dest)) == -1){
			printf("Error sending packet.\n");
			return -1;
		}
		free(packet);
		
		gotpk = 0;
		rpacket = malloc(150);
		start = (u_int32_t)time(NULL);
		while((start+timeout > time(NULL)) && gotpk != 1){
			if((rsize = recv_raw_pf(recvsk, rpacket, 150)) >= 0){
				if((filter_tcp_packet(rpacket, rsize, ip_cfg, tcp_cfg)) ==-1)
					continue;
				else gotpk = 1;
				
				fingerp = extract_fprint(rpacket);
				
				if((search_siglist(fingerp, sigDB, nsigs, out, ip_cfg, 
								   tcp_cfg)) == 0){
					continue;	
				}
		
			}
		
		}
		free(rpacket);
	}
	
	close(sendsk);
	close(recvsk);
	return 0;
}
