/*
 * packet_headers.h -dl
 * Function prototypes and Type definitions for packet building
*/
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "protodefs.h"

/* Packet header config structs*/
typedef struct{
	char *smac;
	char *dmac;
	int ethp;
} linkcfg_t;

typedef struct{
	u_short ip_id;
	u_short frag_off;
	u_short ttl;
	u_short ipp;
	struct in_addr src;
	struct in_addr dest;
} ipcfg_t;

typedef struct{
	u_int32_t seqn;
	u_int32_t ackn;
	u_short sport;
	u_short dport;
	u_short tflags;
	u_short winsz;
	u_short urgp;
	
	char *tcpopts;
	u_short optsz;
} tcpcfg_t;

/* Checksum routine*/
u_short calc_csum(u_char *data, int len);

/* Protocol header build routines */
etherhdr_t *build_ether_hdr(linkcfg_t linkcfg);

iphdr_t *build_ip_hdr(ipcfg_t ipcfg, int trans_sz, int payload_sz);

tcphdr_t *build_tcp_hdr(tcpcfg_t tcpcfg, iphdr_t *ip_hdr, char *payload, 
						int payload_sz);

psuhdr_t *build_pseudo(iphdr_t *ip_hdr);

