/*
 * protodefs.h
 * Mostly from Linux kernel headers
*/
#include <netinet/in.h>

/*====== Link Layer ========================================================*/
#define ETH_ALEN	  	6		/* Bytes in mac address */
#define ETH_HLEN	  	14    	/* Bytes in ethernet header */
#define ETH_DATA_LEN	1500	/* Max bytes in ethernet payload */

#define ETHERTYPE_IP	0x0800
#define ETHERTYPE_ARP   0x0806
#define ETHERTYPE_RARP 	0x0835
#define ETH_P_ALL       0x0003

typedef struct
{
	u_int8_t  ether_dhost[ETH_ALEN];
	u_int8_t  ether_shost[ETH_ALEN];
	u_int16_t ether_type;
} etherhdr_t;



/*====== Network layer =====================================================*/
#define IPVERSION 	  	4
#define IP_MAXPACKET	65535

#define IP_ALEN			4
#define MAXTTL			255
#define IPDEFTTL 		64

typedef struct 
{
#if __BYTE_ORDER == __LITTLE_ENDIAN
	unsigned int ip_hl:4;
	unsigned int ip_v:4;
#endif
#if __BYTE_ORDER == __BIG_ENDIAN
	unsigned int ip_v:4;
	unsigned int ip_hl:4;
#endif
	u_int8_t ip_tos;
	u_short	 ip_len;
	u_short  ip_id;
	u_short  ip_off;
#define IP_RF 		0x0800
#define IP_DF 		0x4000
#define IP_MF 		0x2000
#define IP_OFFMASK 	0x1FFF
	u_int8_t ip_ttl;
	u_int8_t ip_p;
	u_short  ip_sum;
	struct in_addr ip_src, ip_dst;
} iphdr_t;


/*====== TCP ===============================================================*/

typedef struct {
    u_int16_t th_sport;         /* source port */
    u_int16_t th_dport;         /* destination port */
    u_int32_t th_seq;           /* sequence number */
    u_int32_t th_ack;           /* acknowledgement number */
#if __BYTE_ORDER == __LITTLE_ENDIAN
    u_int8_t th_x2:4;           /* (unused) */
    u_int8_t th_off:4;          /* data offset */
#endif
#if __BYTE_ORDER == __BIG_ENDIAN
    u_int8_t th_off:4;          /* data offset */
    u_int8_t th_x2:4;           /* (unused) */
#endif
    u_int8_t th_flags;
#define TH_FIN  0x01
#define TH_SYN  0x02
#define TH_RST  0x04
#define TH_PUSH 0x08
#define TH_ACK  0x10
#define TH_URG  0x20
    u_int16_t th_win;           /* window */
    u_int16_t th_sum;           /* checksum */
    u_int16_t th_urp;           /* urgent pointer */
} tcphdr_t;

# define TCPOPT_EOL             0
# define TCPOPT_NOP             1
# define TCPOPT_MAXSEG          2
# define TCPOLEN_MAXSEG         4
# define TCPOPT_WINDOW          3
# define TCPOLEN_WINDOW         3
# define TCPOPT_SACK_PERMITTED  4               /* Experimental */
# define TCPOLEN_SACK_PERMITTED 2
# define TCPOPT_SACK            5               /* Experimental */
# define TCPOPT_TIMESTAMP       8
# define TCPOLEN_TIMESTAMP      10


/*====== UDP ===============================================================*/
typedef struct
{
	u_int16_t source;
	u_int16_t dest;
	u_int16_t len;
	u_int16_t check;
} udphdr_t;


/*====== Pseudo Header =====================================================*/
typedef struct
{
	struct  in_addr src_ip, dest_ip;
	u_char  reserved;
	u_char  protocol;
	u_short len;
} psuhdr_t;

