/*
 * utility.c  -dl
 * General Utility Functions for Inth
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


/*
 * Calculate the difference between two timeval structure's
*/
void time_diff(struct timeval *start, struct timeval *fin)
{
	int temp;

	/* Calculate seconds */
	if(fin->tv_sec >= start->tv_sec)
        temp = (fin->tv_sec - start->tv_sec);

	/* Calculate milliseconds */
	if(fin->tv_usec < start->tv_usec){
		fin->tv_usec =  (fin->tv_usec - start->tv_usec) + 1000000;
		if(temp >= 1)
			fin->tv_sec = temp - 1;
	}
	else{
		fin->tv_usec = (fin->tv_usec - start->tv_usec);
		fin->tv_sec = temp;
	}
}

/*
 * allocates and zeros out memory of size, returns pointer to memory
 * area on success, or exit -1 on failure
*/
void *xmalloc(size_t size)
{
	register void *ptr;

	if(!(ptr = malloc(size))){
    	printf("malloc(): Unable to allocate memory!\n");
		exit(-1);
	}

	memset(ptr, 0, size);
	return ptr;
}

/*
 * rellocates and zeros out memory of size, returns pointer to new memory
 * area on success, or exit -1 on failure
*/
void *xrealloc(void *ptr, size_t size)
{
	register void *nptr;

	if(!(nptr = realloc(ptr, size))){
		printf("realloc(): Unable to allocate memory!\n");
    	exit(-1);
	}

	memset(nptr, 0, size);
	return nptr;
}

/*
 * Handle ports inputted at the command line in comma or hyphen seperated 
 * format, like so '21,22,23,25' or '21-113' or both '21,22,23-25', returns
 * Number of ports in array on success or -1 on failure. (needs to be redone)
*/
int handle_ports(char *pstring, u_short *ports)
{
	char c, *tmp;
	int x, i = 0, a, b, d = 0, sum = 0,psum=0;

	/* If only one port, return */

	/* Count memory needed */
	x = 0;
	while((c = pstring[x++]) != '\0'){
		if(c == '-'){
			a = 0, b = 0;
			if((sscanf(pstring, "%d-%d", &a, &b)) != 2){
				printf("FIXME:Cant parse hyphen if commas are first!!\n");
				return -1;
			}else{
				if(b > a && a > 0) sum = b - a;
				if(a > b && b > 0) sum = a - b;
			}
		}
	}
	
	x = 0;
	while((c = pstring[x++]) != '\0'){
		if(c == ',' && i != 1){
			a = 0, b = 0;
			while((c = pstring[a++]) != '\0'){
				if(c == ',') b++;
			}
			if(i == 0) i++;
			else x = strlen(pstring);
			sum += b++;
		}
	}
	
	/* Allocate memory and initialize ports list */
	/* Do i Need to declare this locally and have a */
	/* (u_short *)return?? For now this is fine unless */
	/* it causes problems or breaks (allocated before call' */
	/* with xmalloc(sizeof(u_short))' */
	ports = xrealloc(ports, (sum+1) * sizeof(u_short));
	
	/* Now to fill out *ports list */
	x = 0;
	while((c = pstring[x++]) != '\0'){
		if(c == '-'){
			a = 0, b = 0;
			if((sscanf(pstring, "%d-%d", &a, &b)) != 2){
				printf("FIXME:Cant parse hyphen if commas are first!!\n");
				return -1;
			}else{
				if(b > a && a > 0){
					for(d = a; d <= b; d++)
						ports[psum++] = d;
				}
				else if(a > b && b > 0){
					for(d = b; d <= a; d++)
						ports[psum++] = d;
				}
			}
		}
	}
	
	x = 0;
	while((c = pstring[x++]) != '\0'){
		if(c == ','){
			tmp = strtok(pstring, ",");
			ports[psum++] = atoi(tmp);
			while((tmp = strtok(NULL, ",")) != NULL){
				ports[psum++] = atoi(tmp);
			}	
		}
	}
	
	/* DBG
	for(i = 0; i < psum; i++)
		printf("%d\n",ports[i]); */
	
	
	return psum;
}

/*
 * Acts like the perl chomp() function removing the new line 
 * from a string and null terminating it
*/
void chomp(char *string)
{
	int len = strlen(string);
	if(string[len-1] == '\n')
		string[len-1] = '\0';
}
